@@ -1,6 +1,103 @@
-# $Header: /usr/local/src/cvsroot/tcpquota/lib/tcpquota.pl,v 1.39 1998/04/16 18:45:43 turbo Exp $
+# $Header: /usr/lib/cvs/root/tcpquota/lib/tcpquota.pl,v 1.55 1998/08/02 17:57:50 turbo Exp $
 #
 # $Log: tcpquota.pl,v $
+# Revision 1.55  1998/08/02 17:57:50  turbo
+# * Removed the PORT variable, the port is in the SERVER variable...
+# * Better spaced debug output when loading the config file, also a 'done'
+#   line at the end...
+# * On papadoc, when I'm running screen/X, I get the host address ':0:S.0',
+#   instead of 'papadoc.nocrew.org:0:S.0'... Take this into account, hopfully
+#   it woun't break any thing else...
+# * Try to find the address to the 'ut_host' instead of the 'ut_addr'... Hmmm..
+#   Wonder how/why i tried to find a address, when I already have one... :)
+#
+# Revision 1.54  1998/08/01 19:59:41  turbo
+# * First quick port to use the generic database interface 'DBI' instead of
+#   the 'Msql' interface. This is so that we can go from using mSQL as
+#   database, to use the much quicker mySQL server. But by using this generic
+#   interface, we can have both... More or less :)
+# # Any reference to the Msql function 'query' had to be replaced with, first
+#   a 'prepare' then a 'execute'. If the execute fails, then die, or log, or
+#   what evere takes us fancy...
+# * Any reference to the Msql functions 'fetchrow' and 'numrows', had to be
+#   replaced with 'fetchrow_array' and 'rows'.
+# * Any reference to the variable 'dh' (as in DatabaseHandle) and 'quotadbh',
+#   have been replaced with 'dbh' all over the board, so that we have one common
+#   denominator.
+# * Found a 'open_sql_server()' function in the 'tcp_masq_openfw' script. Move
+#   that to the library, so that we can reuse the function all over the board.
+#
+# Revision 1.53  1998/06/12 14:47:23  turbo
+# * Logg username/hostname if we are running in debug mode...
+# * Spell errors fixed in the english translation... I have to rewrite this
+#   some day... *sigh*
+#
+# Revision 1.52  1998/05/29 14:16:49  turbo
+# Check the last three minutes worth of masquerading connects instead of the
+# last PERIOD sekonds...
+#
+# Revision 1.51  1998/05/25 16:09:31  turbo
+# * Print a bar '|', to separate the two columns in the list database.
+# * Do not try to calculate the cost from the LOW_RATE value.... We only need to
+#   return number of minutes...
+#
+# Revision 1.50  1998/05/24 15:48:08  turbo
+# * Print the quota output in TWO columns, instead of one (chktcpquota -a)
+#   Function: 'list_dbase()'.
+#
+# Revision 1.49  1998/05/17 11:12:13  turbo
+# Bug fixing 'check_localnet()'. Make sure we check each localnet with the
+# input to the function (network we want to no if local).
+#
+# Revision 1.48  1998/05/14 16:57:09  turbo
+# * Added two new functions:
+#     find_basenet()           => Convert a '42.42.0.0' net to '42.42'.
+#     check_localnet()         => Check if the network is one of the local ones.
+# * Check for a local network at the same time as we are loading the network
+#   trafic data.
+# * When waiting for the daemon to close the firewall, when we get the 'closed'
+#   from the server, make sure to delete the entry from the database...
+#
+# Revision 1.47  1998/04/29 13:48:48  turbo
+# If we can not find the IP number to the host, return 0 (not -1), and output
+# a nice, and descriptive explanation to what's happening...
+#
+# Revision 1.46  1998/04/26 20:10:47  turbo
+# Instead of hardcoding the utmp entry number, use the function's
+# that is provided... (ex: '&utmp'ut_addr' returns number 7, the
+# sevent'th entry in the utmp struct).
+# The only thing one have to do for a libc6 system is to include
+# another 'utmp.ph' file (which is from a libc6 utmp.h).
+#
+# Revision 1.45  1998/04/26 14:37:50  turbo
+# Fix the problem with short addresses, like: '205.229.104.5' (oct/hex: 90760653/568e5cd)
+# Hopfully it fixes other problems to... :)
+#
+# Revision 1.44  1998/04/19 19:51:10  turbo
+# When we try to exchange a '1' to true, don't!
+# Some values _IS_ numeric... (Boolean is now
+# ether 'true/yes' or 'false/no').
+#
+# Revision 1.43  1998/04/19 09:30:46  turbo
+# Added real support for a boolean value.
+# If a value is 1 or true, it should be changed
+# to 'yes', and vise versa...
+#
+# Revision 1.42  1998/04/18 18:19:08  turbo
+# * No need to do a 'foreach' in the message, just print the whole
+#   thing at once...
+# * Security fix in the 'open()'...
+#
+# Revision 1.41  1998/04/18 16:38:36  turbo
+# Added a new function, send_message, which sends a specific message
+# to a user. It first sends to there controlling terminal, and if
+# possible to there WINPOPUP.EXE if they are running windows (via
+# one of the SAMBA programs)...
+#
+# Revision 1.40  1998/04/16 20:20:47  turbo
+# If the IP number can't be found int the utmp file,
+# look up the hostname instead...
+#
 # Revision 1.39  1998/04/16 18:45:43  turbo
 # * Don't fetch the whole database table when we know that we are waiting for one
 #   specific host... Just select the one host we are waiting for!
@@ -187,7 +284,7 @@ sub readconfig {
 	chop($tmp);
 
 	# Skip comments and empty lines...
-	next if( $tmp =~ /^#/ );
+	next if( $tmp =~ /^\#/ );
 	next if( $tmp =~ /^$/ );
 
 	# local config for this prog.  remote proceding string.
@@ -196,13 +293,30 @@ sub readconfig {
 	}
 
 	($name,$var)=split(/=/, $tmp);
-	print "$name=$var\n" if $DEBUG;
+
+	$var = 'yes' if( $var eq 'true' );
+	$var = 'no'  if( $var eq 'false' );
+
+	printf("  %-24s = %s\n", $name, $var) if $DEBUG;
 	$cf{$name} = $var;
     }
 
+    print "Done reading config...\n\n" if $DEBUG;
     close(CF);
 }
 
+######################################################################
+#
+# Open the SQL server...
+#
+sub open_sql_server {
+    undef( $dbh );
+
+    # Open up the database connection...
+    $dbh = DBI->connect( "dbi:$cf{'ENGINE'}:tcpquota:$cf{'SERVER'}")
+	|| die( "Can't connect to database at '$cf{'SERVER'}." );
+}
+
 ######################################################################
 #
 #
@@ -210,11 +324,9 @@ sub logg_actions {
     local($database, $string) = @_;
     my($query, $sth);
 
-    # Keep quiet when sql errors.
-    local($Msql::QUIET) = 1;
-
     # Make sure the table exists...
-    if(! $dbh->query("select * from $database") ) {
+    $sth = $dbh->prepare("select * from $database");
+    if(! $sth->execute ) {
 	# Table does not exists, create it...
 	$query  = "create table $database ( ";
 	$query .=   "when char(14) not null, ";
@@ -222,15 +334,15 @@ sub logg_actions {
 	$query .=   "action char(57) not null ";
 	$query .= ")";
 
-	$sth = $dbh->query($query)
-	    || die( "Error in creating table logging...\n" );
+	$sth = $dbh->prepare($query);
+	$sth->execute || die "Could not execute query: $sth->errstr";
     }
 
     $date = get_timestring();
 
     $query = "insert into $database (when,user,action) values ('$date', '$USER', '$string')";
-    $sth = $dbh->query($query)
-	|| die( "Error in inserting in the logging table...\n" );
+    $sth = $dbh->prepare($query);
+    $sth->execute || die "Could not execute query: $sth->errstr";
 }
 
 ######################################################################
@@ -243,20 +355,17 @@ sub list_logg {
     local($database, $username) = @_;
     my($date, $user, $action, $query, $sth);
 
-    # Keep quiet when sql errors.
-    local($Msql::QUIET) = 1;
-
     # Query the database.
     if( $username ) {
 	$query = "select * from $database where user like '$username'";
     } else {
 	$query = "select * from $database";
     }
-    $sth = $dbh->query($query)
-	|| die("There is no logging so faar...\n");
+    $sth = $dbh->prepare($query);
+    $sth->execute || die("There is no logging so faar...\n");
 
     printf("Date     Time  User       Action\n");
-    while( ($date,$user,$action) = $sth->fetchrow ) {
+    while( ($date,$user,$action) = $sth->fetchrow_array ) {
 	printf("%-14s %-10s %-40s\n", $date, $user, $action);
     }
 
@@ -271,26 +380,33 @@ sub list_logg {
 #
 sub list_dbase {
     local($dbase) = @_;
-    my(%line, $cost, $sec, $user, $key, $sth);
+    my(%line, $cost, $sec, $user, $key, $sth, $i);
 
     # Query the database.
-    $sth = $dbh->query("select * from $dbase")
-	|| terminate( "Error when query..." );
+    $sth = $dbh->prepare("select * from $dbase");
+    $sth->execute || &terminate( "Error when query..." );
 
     if( $dbase eq 'tcptab' ) {
-	while( ($user,$sec) = $sth->fetchrow ) {
+	while( ($user,$sec) = $sth->fetchrow_array ) {
 	    $cost = calculate_cost($sec);
 	    $line{$user} = sprintf("%-10s %9d  (%5d kronor)", $user, $sec, $cost);
 	}
     } else {
-	while( $user = $sth->fetchrow ) {
+	while( $user = $sth->fetchrow_array ) {
 	    $line{$user} = $user;
 	}
     }
 
     # Sort the list...
+    $i = 0;
     foreach $key (sort(keys %line)) {
-	printf("$line{$key}\n");
+	if(! $i ) {
+	    printf("$line{$key}\t|  ");
+	    $i++;
+	} else {
+	    printf("$line{$key}\n");
+	    $i--;
+	}
     }
 
     print "\n";
@@ -352,6 +468,7 @@ sub info {
 sub calculate_cost {
     local($sec) = @_;
     my($add_minus) = 0;
+    my($b);
 
     if(! $sec ) {
 	$sec = 0;
@@ -364,15 +481,12 @@ sub calculate_cost {
     # calculate number of minutes.
     $b = int( ($sec / 600) );
 
-    # calculate the cost.
-    $c = $b * $cf{'LOW_RATE'};
-
     if( $add_minus ) {
 	# Add the '-' again...
-	$c = "-$c";
+	$b = "-$b";
     }
 
-    return $c;
+    return $b;
 }
 
 ######################################################################
@@ -406,7 +520,7 @@ sub find_ip {
 	($a,$b,$c,$d) = unpack('C4', $addr);
 	return("$a.$b.$c.$d");
     } else {
-	return(-1);
+	return(0);
     }
 }
 
@@ -438,6 +552,69 @@ sub find_fqdn {
     return( $fqdn );
 }
 
+######################################################################
+#
+# proc: find_basenet( void )
+#
+# get the base network address
+# 42.42.0.0  => 42.42
+# 
+sub find_basenet {
+    my(@locnet, @temp, %network, $tmp, $i);
+
+    # Cycle through the localnets...
+    @locnet = split(' ', $localnet);
+    foreach $tmp (@locnet) {
+	@temp = split('\.', $tmp);
+	$network{$tmp} = "";
+
+	# Convert 42.42.0.0 => 42.42
+	for($i = 0; $i <= 3; $i++) {
+	    if($temp[$i]) {
+		# Remove the net part...
+		$network{$tmp} .= $temp[$i];
+	    }
+
+	    if($temp[$i+1]) {
+		$network{$tmp} .= ".";
+	    }
+	}
+    }
+
+    return(%network);
+}
+
+sub check_localnet {
+    local($src, %network) = @_;
+    my($tmp, $local_network, $out);
+
+    $local_network = 0; $out = "CHK: ";
+  NET:
+    foreach $tmp (keys %network) {
+	# Skip connect with our own network addresses...
+
+	$out .= sprintf("$src ($network{$tmp}) ");
+	if( $src =~ /^$network{$tmp}/ ) {
+	    $out .= sprintf("=> LOCAL!\n");
+	    $local_network = 1;
+
+	    last NET;
+	}
+    }
+
+    if( $src eq $cf{'LOC_ADDR'} ) {
+	$out .= sprintf("=> LOCAL!\n");
+	$local_network = 1;
+    }
+
+    if(! $local_network) {
+	$out .= sprintf("=> ! LOCAL\n");
+    }
+
+    &logg(2, $out);
+    return($local_network);
+}
+
 ######################################################################
 #
 # proc: find_users()
@@ -445,7 +622,7 @@ sub find_fqdn {
 # Get the users currently online...
 #
 sub get_online {
-    my($i, $a, $b, $c, $d, $utmp, $addr);
+    my($i, $a, $b, $c, $d, $utmp, $addr, $logg_string);
     my(%ret) = ();
 
     open(UTMP, '/var/run/utmp');
@@ -454,32 +631,45 @@ sub get_online {
     while(read(UTMP, $utmp, &utmp'sizeof())) {
         @utmp = unpack(&utmp'typedef(), $utmp);
 
-        if($utmp[5]) {
-            next if($utmp[5] =~ /^LOGIN/);
-            next if($utmp[6] =~ /^0/);
-            next if($utmp[0] == 8);
+        if($utmp[&utmp'ut_user]) {
+            next if($utmp[&utmp'ut_user] =~ /^LOGIN/);
+            next if($utmp[&utmp'ut_host] =~ /^0/);
+            next if($utmp[&utmp'ut_type] == 8);
 
             # Get the hostname, but allow ':x.x' at the end of the hostname string...
 	    # Converts 'stalin.ccw.se:1.0' to 'stalin.ccw.se'...
-	    $utmp[6] =~ /([^:]*).*/;
-	    $utmp[6] = $1;
+	    $utmp[&utmp'ut_host] =~ /([^:]*).*/;
+	    $utmp[&utmp'ut_host] = $1;
+
+	    # Just incase we have a ':0:S.0' address, the ut_host above will be empty...
+	    if(! $utmp[&utmp'ut_host] ) {
+                $utmp[&utmp'ut_host] = $cf{'LOC_ADDR'};
+	    }
 
 	    # Do we have a source address?
-	    if( $utmp[7] ) {
+	    if( $utmp[&utmp'ut_addr] ) {
 		# Octal value == Hex value == IP address (reversed)
 		# 11512581674 == 5A282A2A  == 90.40.42.42
 
 		# Get it as a hex digit.
-		$src = sprintf("%x", $utmp[7]);
+		$src = sprintf("%x", $utmp[&utmp'ut_addr]);
+		if(length($src) == 7) {
+		    # Fix the problem with the address: '205.229.104.5' (oct/hex: 90760653/568e5cd)
+		    # Hopfully it fixes other problems to... :)
+		    $src = "0" . $src;
+		}
 
 		# Now, output the IP address, in propper order...
 		$addr = &revert_ip(&hex_to_ip($src));
 	    } else {
-		$addr = $utmp[7];
+		$addr = &find_ip($utmp[&utmp'ut_host]);
 	    }
 
 	    # Compose: username hostname tty IPaddr
-	    $ret{$i} = sprintf("%-10s %-35s %-7s $addr", $utmp[5], $utmp[6], $utmp[2]);
+	    $ret{$i} = sprintf("%-10s %-35s %-7s $addr", $utmp[&utmp'ut_user], $utmp[&utmp'ut_host], $utmp[&utmp'ut_line]);
+
+	    $logg_string = "WHO: $utmp[&utmp'ut_user]/$utmp[&utmp'ut_host]";
+	    &logg(2, "$logg_string\n") if($DEBUG);
 
             $i++;
         }
@@ -496,22 +686,16 @@ sub get_online {
 # Get the connect to and from us, not counting any local comm...
 #
 sub get_netstat {
-    my(%netstat, @temp, $line, $i, $src, $dst, $source, $destin, $uid, $net);
+    my(%netstat, %network);
+    my($destin, $uid, $line, $src, $dst, $i);
+    my($source, $local_network);
 
-    $net = "";
-    @temp = split('\.', $localnet);
-    for($i = 0; $i < 4; $i++) {
-	if($temp[$i]) {
-	    $net .= $temp[$i];
-	}
-
-	if($temp[$i+1]) {
-	    $net .= ".";
-	}
-    }
+    %network = &find_basenet();
 
     # Read from 'netstat'...
-    open(NETSTAT, "/proc/net/tcp") || &logg(1,"ERROR: Could not open /proc/net/tcp, $!");
+    open(NETSTAT, "/proc/net/tcp")
+	|| &logg(1,"ERROR: Could not open /proc/net/tcp, $!");
+
     <NETSTAT>; $i = 0;
     while(! eof(NETSTAT) ) {
 	$line = <NETSTAT>;
@@ -529,7 +713,7 @@ sub get_netstat {
 	$source = &revert_ip(&hex_to_ip((split(':', $src))[0])) . ":" . hex((split(':', $src))[1]);
 	$destin = &revert_ip(&hex_to_ip((split(':', $dst))[0])) . ":" . hex((split(':', $dst))[1]);
 
-	next if( $source =~ /^$net/ );  # Skip connect with our own network address
+	next if( &check_localnet($destin, &find_basenet()) );
 
 	$netstat{$i} = "$source $destin $uid $stat";
 	$i++;
@@ -593,7 +777,8 @@ sub get_masqueraded_online {
 	    ($minutes, $time)    = split(/:/,  $time);
 	    ($seconds, $milisec) = split(/\./, $time);
 
-	    if( $minutes >= 14 && ($seconds >= 20 - $cf{'PERIOD'}) ) {
+	    # Wrong.... We take the last three minutes...
+	    if( $minutes >= 12 && ($seconds >= 20 - $cf{'PERIOD'}) ) {
 		$cnts++;
 	    }
 	}
@@ -777,11 +962,11 @@ sub wait_for_masq {
     local($ip, $command) = @_;
     my($sth, $host, $name, $cnts, $tic, $count,$free,$open);
 
-    $sth = $dbh->query("select * from masq where host like '$ip'") || 
-	die( "SELECT: Could not fetch from masq table, $!\n" );
+    $sth = $dbh->prepare("select * from masq where host like '$ip'");
+    $sth->execute || die( "SELECT: Could not fetch from masq table: $sth->errstr" );
 
     # Fetch each hit..
-    ($host, $name, $cnts, $tic, $count, $open, $free) = $sth->fetchrow;
+    ($host, $name, $cnts, $tic, $count, $open, $free) = $sth->fetchrow_array;
 
     if( $host ) {
 	# This is the host/ip we are waiting for, is it opened/closed?
@@ -789,17 +974,89 @@ sub wait_for_masq {
 	    # It is opened...
 	    return( 1 );
 	} elsif( ($command eq 'close') && ($open eq '0') ) {
-	    # It is closed...
+	    # Remove the entry from the database...
+	    $sth = $dbh->prepare("delete from masq where host = '$ip'");
+	    $sth->execute || die "Sorry could not delete from masq: $dbh->errstr";
+
 	    return( 1 );
 	} else {
 	    return( 0 );
 	}
     } else {
-	# Could not fetch the line in the table, tell the caller that it's okay...
+	# Could not fetch the line in the table, tell the caller that it's okay anyway...
 	return( 1 );
     }
 }
 
+#
+# Send a message to a specified user
+# Both on there controlling terminal, and if possible to there
+# WINPOPUP.EXE if they are running windows (via one of the SAMBA
+# programs)...
+sub send_message {
+    local($user, @message) = @_;
+    my(%WHO, $tmp, $dummy, $username);
+
+    # Do a write to the user's terminal...
+    if( open(WRITE,"|/usr/bin/write $user") ) {
+	# Send the message...
+	print WRITE @message;
+
+	# Close the pipe...
+	close(WRITE);
+    } else {
+	&logg(1, "Could not open '/usr/bin/write $user', $!\n");
+    }
+
+    # First find the host the user is comming from...
+    %WHO = &get_online();
+
+    $i = 0;
+    while( ($username, $host, $dummy, $dummy) = split(' ', $WHO{$i}) ) {
+	if( $username eq $user ) {
+	    # Untaint username, to make the command line secure...
+	    if( $host =~ /^([-\@\w.]+)$/) {
+		$host = $1;
+	    } else {
+		&logg(1, "Can not send a SMB messge to $user\@$host, not taint secure\n");
+		return;
+	    }
+
+	    $ip = &find_ip($host);
+	    # Try to open the smbclient program...
+	    if( open( SMBCLIENT, "|/usr/bin/smbclient -M $host -I $ip > /dev/null" ) ) {
+		# We have found our user, try to send a windows message...
+		# The client have to be running 'winpopup' for this to work...
+		print SMBCLIENT @message;
+
+		# Send a short generic message, like a 'signature'...
+		if( $cf{'LANGUAGE'} eq 'svenska' ) {
+		    print SMBCLIENT "\n\nFr att krympa ner WINPOPUP (eller om du kr\n";
+		    print SMBCLIENT "WinPopup PLUS), och ta bort detta meddelande,\n";
+		    print SMBCLIENT "tryck p papperskorgen (eller 'Close' hgst\n";
+		    print SMBCLIENT "upp till vnster...\n\n";
+		    print SMBCLIENT "Med vnliga hlsningar, administratren.\n";
+		} else {
+		    print SMBCLIENT "\n\nTo shrink this window, and remove the message,\n";
+		    print SMBCLIENT "press the trashcan at the top left...\n";
+		    print SMBCLIENT "Regards, the administrator.\n";
+		}
+
+		# Close the pipe...
+		close(SMBCLIENT);
+
+		# We are done, return...
+		return;
+	    } else {
+		&logg(1, "Could not send a SMB messge to $user\@$host, $!\n");
+		return;
+	    }
+	}
+
+	$i++;
+    }
+}
+
 #-----------------------------------------------------------------
 #
 # Msql funcs.. Dessa meckar med msql och interfacet mot tcpquotad
@@ -815,7 +1072,7 @@ sub closedown {
 # Skriv till msql, s att tcpquotad kan rkna med oss..
 sub write_db {
     local($action, $db, $ip, $name, $verbose, $wait) = @_;
-    local($sth);
+    my($sth, $rv, $regname);
     
     if( $action eq 'add' ) {
 	# Add host to masquerading table...
@@ -825,39 +1082,41 @@ sub write_db {
 	    $free = 0;
 	}
 
-	$sth = $db->query("insert into masq (host,name,cnts,tic,counter,open,free)
-                     values('$ip','$name',0,$cf{TICS},0,2,$free)");
-	if(!$sth) {
-	    if ($db->errmsg() =~ /^Non unique key value in field/) {
-		local($sth,$regname);
-
-		$sth = $db->query("select name from masq where host = '$ip'");
-		$regname = $sth->fetchrow();
-		if($regname ne $name) {
-		    if(! &check_free_user($regname) ) {
-			if( $cf{'LANGUAGE'} eq 'svenska' ) {
-			    die "$regname har ppnat $ip fr internet access enl mina anteckningar.\nNgot r fel, tala med ngon ansvarig...\n";
-			} else {
-			    die "$regname have opened $ip for internet access according to my notes.\nSomething is wrong, speak to someone in charge...\n";
-			}
-		    } else {
-			# This host is opened for free...
-			if( $cf{'LANGUAGE'} eq 'svenska' ) {
-			    die "Datorn ($ip) ar redan oppnad for fri internet access enligt mina anteckningar.\nDet innebar att du inte behover oppna den sjalv...\n";
-			} else {
-			    die "This host ($ip) is opened for free internet acces according to my notes.\nThis means that you don't have to openit yourself...\n";
-			}
-		    }
-		} else  {
+	$sth = $dbh->prepare("select * from masq where host like '$ip'");
+	$rv  = $sth->execute;
+	if( $rv ne '1' ) {
+	    # There is already an entry with this IP in the database...
+	    $sth = $dbh->prepare("select name from masq where host = '$ip'");
+	    $regname = $sth->fetchrow_array;
+	    if($regname ne $name) {
+		if(! &check_free_user($regname) ) {
 		    if( $cf{'LANGUAGE'} eq 'svenska' ) {
-			die "Du har ju redan ppna denna maskin fr internet access.\nKolla att du inte har flera telnet sessioner igng!\n";
+			die "$regname har ppnat $ip fr internet access enl mina anteckningar.\nNgot r fel, tala med ngon ansvarig...\n";
 		    } else {
-			die "Your have already opened this machine for internet access.\nDouble check to se that you do not have any more telnet session running!\n";
+			die "$regname have opened $ip for internet access according to my notes.\nSomething is wrong, speak to someone in charge...\n";
+		    }
+		} else {
+		    # This host is opened for free...
+		    if( $cf{'LANGUAGE'} eq 'svenska' ) {
+			die "Datorn ($ip) ar redan oppnad for fri internet access enligt mina anteckningar.\nDet innebar att du inte behover oppna den sjalv...\n";
+		    } else {
+			die "This host ($ip) is opened for free internet access according to my notes.\nThis means that you don't have to openit yourself...\n";
 		    }
 		}
+	    } else  {
+		if( $cf{'LANGUAGE'} eq 'svenska' ) {
+		    die "Du har ju redan ppna denna maskin fr internet access.\nKolla att du inte har flera telnet sessioner igng!\n";
+		} else {
+		    die "Your have already opened this machine for internet access.\nDouble check to see that you do not have any more telnet session running!\n";
+		}
 	    }
-	    die "Could not update the mSQL database.\n Got:".$db->errmsg()."\n";
+	} else {
+	    # No entry with this IP in the database...
+	    $sth = $dbh->prepare("insert into masq (host,name,cnts,tic,counter,open,free)
+                     values('$ip','$name',0,$cf{TICS},0,2,$free)");
+	    $sth->execute || die "Could not insert into the masq table: $sth->errstr";
 	}
+
 	$db_masq_written=1;
 	print "\n" if($verbose);
 
@@ -884,7 +1143,7 @@ sub write_db {
 		print "You will now be debited for IP traffic to $ip.\n";
 	    }
 	}
-    } else {
+    } elsif( $action eq 'rem' ) {
 	# Remove host from masquerade table...
 
 	if($verbose) {
@@ -896,10 +1155,8 @@ sub write_db {
 	}
 
 	# Tell the database/daemon we want this host closed...
-	$sth = $db->query("update masq set open = 3 where host = '$ip'");
-	if($db->errmsg()) {
-	    die "Sorry, got ".$db->errmsg."\n";
-	}
+	$sth = $dbh->prepare("update masq set open = 3 where host = '$ip'");
+	$sth->execute || die "Could not execute query: $sth->errstr";
 	$db_masq_written=0;
 
 	if($wait) {
@@ -909,10 +1166,8 @@ sub write_db {
 	    }
 
 	    # Remove the entry from the database...
-	    $sth = $db->query("delete from masq where host = '$ip'");
-	    if($db->errmsg()) {
-		die "Sorry, got ".$db->errmsg."\n";
-	    }
+	    $sth = $dbh->prepare("delete from masq where host = '$ip'");
+	    $sth->execute || die "Could not execute query: $sth->errstr";
 	    $db_masq_written=0;
 	}
 
